
# THIS FILE IS GENERATED FROM SCIPY SETUP.PY
short_version = '0.1.10'
version = '0.1.10'
full_version = '0.1.10'
git_revision = '41f20c35c008a20793efa7fa63f3a7233a53c63c'
release = True
if not release:
    version = full_version
